// **********************************************************************
//
// Copyright (c) 2005-2006
// IONA Technologies PLC
// Dublin, Ireland and Waltham, MA, USA
//
// All Rights Reserved
//
// Licensed to apache under Apache 2.0 License - 2006
// **********************************************************************

package com.ooc.OB;

final class GIOPServerStarterThreaded extends GIOPServerStarter
{
    //
    // The starter thread
    //
    protected final class StarterThread extends Thread
    {
	private GIOPServerStarterThreaded starter_;
	
        StarterThread(ThreadGroup group, GIOPServerStarterThreaded starter)
        {
	    super(group, "ORBacus:Server:StarterThread");
	    starter_ = starter;
        }
        
	public void
        run()
	{
            try
            {
                starter_.starterRun();
            }
            catch(RuntimeException ex)
            {
                ex.printStackTrace();
                Assert._OB_assert(false);
            }

            //
            // Shutdown the acceptor so that no further connections are
            // accepted
            //
            starter_.logCloseAcceptor();
            starter_.acceptor_.shutdown();

            //
            // Accept all connections which might have queued up in the
            // listen() backlog
            //
            while(true)
            {
                com.ooc.OCI.Transport transport = null;

                try
                {
                    transport = starter_.acceptor_.accept(false);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                }

                if(transport == null)
                    break;

                try
                {
                    GIOPConnection connection =
                        new GIOPConnectionThreaded(starter_.orbInstance_,
					           transport,
					           starter_.oaInterface_);

                    connection.setState(GIOPConnection.State.Closing);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    // Ignore SystemExceptions
                }
            }

            //
            // Close the acceptor
            //
            starter_.acceptor_.close();

            //
            // Break cyclic object dependency
	    //
	    starter_ = null;
 	}
    }
    protected Thread starterThread_;

    // ----------------------------------------------------------------------
    // GIOPServerStarterThreaded package member implementation
    // ----------------------------------------------------------------------

    GIOPServerStarterThreaded(ORBInstance orbInstance,
                              com.ooc.OCI.Acceptor acceptor,
			      OAInterface oaInterface)
    {
	super(orbInstance, acceptor, oaInterface);

        try
        {
            //
            // Retrieve the thread group for the servers
            //
            ThreadGroup group = orbInstance_.getServerWorkerGroup();

            //
            // Start starter thread
            //
            starterThread_ = new StarterThread(group, this);
            starterThread_.start();
        }
        catch(OutOfMemoryError ex)
        {
            acceptor_.close();
            state_ = StateClosed;
            throw new org.omg.CORBA.IMP_LIMIT(
                MinorCodes.describeImpLimit(MinorCodes.MinorThreadLimit),
                MinorCodes.MinorThreadLimit,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    // ----------------------------------------------------------------------
    // GIOPServerStarterThreaded public member implementation
    // ----------------------------------------------------------------------

    //
    // Change the state of the worker
    //
    synchronized public void
    setState(int state)
    {
        //
        // Don't do anything if there is no state change
        //
        if(state_ == state)
            return;

        //
        // It is not possible to transition backwards, except if we are
        // in holding state
        //
        if(state_ != StateHolding && state < state_)
            return;

        switch(state)
        {
        case StateActive:
	{
	    for(int i = 0 ; i < connections_.size() ; i++)
	    {
	        GIOPConnection w = (GIOPConnection)connections_.elementAt(i);
	        w.setState(GIOPConnection.State.Active);
	    }
	    
	    break;
	}

        case StateHolding:
	    for(int i = 0 ; i < connections_.size() ; i++)
	    {
	        GIOPConnection w = (GIOPConnection)connections_.elementAt(i);
	        w.setState(GIOPConnection.State.Holding);
	    }
            break;

        case StateClosed:
	{
            for(int i = 0 ; i < connections_.size() ; i++)
            {
                GIOPConnection w = (GIOPConnection)connections_.elementAt(i);
                w.setState(GIOPConnection.State.Closing);
            }
            connections_.removeAllElements();

            //
            // Connect to this starter's acceptor, to unblock the call
            // to accept() in the starter thread
            //
            try
            {
                com.ooc.OCI.Transport tr = acceptor_.connect_self();
                tr.close();
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                //
                // Ignore all system exceptions
                //
            }

            break;
	}
        }

        //
        // Update the state and notify about the state change
        //
        state_ = state;
        notifyAll();
    }

    //
    // Run method for starter thread
    //
    public void
    starterRun()
    {
        while(true)
        {
            //
            // Get new transport, blocking
            //
            com.ooc.OCI.Transport transport = null;
            try
            {
                transport = acceptor_.accept(true);
                Assert._OB_assert(transport != null);
            }
            catch(org.omg.CORBA.NO_PERMISSION ex)
            {
                //
                // Ignore NO_PERMISSION exceptions
                //
            }
            catch(org.omg.CORBA.SystemException ex)
            {
                //
                // Ignore exception. This probably means that the server
                // exceeded the number of available file descriptors.
                //
            }


            synchronized(this)
            {
	        //
		// Reap the existing set of workers
		//
		reapWorkers();

                //
                // Check whether we are on hold
                //
                while(state_ == StateHolding)
                {
                    try
                    {
                        wait();
                    }
                    catch(InterruptedException ex)
                    {
                    }
                }

                if(transport != null)
                {
                    try
                    {
                        if(state_ == StateActive)
                        {
                            //
                            // If we're active, we create and add a new
                            // worker to the worker list
                            //
                            GIOPConnection connection =
                                new GIOPConnectionThreaded(orbInstance_,
						           transport,
						           oaInterface_);
                            connections_.addElement(connection);
                            connection.setState(GIOPConnection.State.Active);
                        }
                        else
                        {
                            //
                            // If we're closed, we create a new dummy
                            // worker, only in order to set it to
                            // StateClosing for proper connection shutdown
                            //
                            Assert._OB_assert(state_ == StateClosed);

                            GIOPConnection connection =
                                new GIOPConnectionThreaded(orbInstance_,
						           transport,
						           oaInterface_);
			    
                            connection.setState(GIOPConnection.State.Closing);
                        }
                    }
                    catch(org.omg.CORBA.SystemException ex)
                    {
                        String msg = "can't accept connection\n" +
                            ex.getMessage();
                        orbInstance_.getLogger().warning(msg);
                    }
                }

                if(state_ == StateClosed)
                    break;
            }
        }
    }
}
