// **********************************************************************
//
// Copyright (c) 2005-2006
// IONA Technologies PLC
// Dublin, Ireland and Waltham, MA, USA
//
// All Rights Reserved
//
// Licensed to apache under Apache 2.0 License - 2006
// **********************************************************************

package com.ooc.OB;

final public class ValueReader
{
    //
    // Chunk data
    //
    static private class ChunkState
    {
        boolean chunked;
        int nestingLevel;
        int chunkStart;
        int chunkSize;

        ChunkState()
        {
        }

        ChunkState(ChunkState s)
        {
            copyFrom(s);
        }

        void
        copyFrom(ChunkState s)
        {
            chunked = s.chunked;
            nestingLevel = s.nestingLevel;
            chunkStart = s.chunkStart;
            chunkSize = s.chunkSize;
        }
    }

    //
    // Valuetype header data
    //
    static private class Header
    {
        int tag;
        int headerPos;
        int dataPos;
        String[] ids;
        ChunkState state;
        Header next; // Java only

        Header()
        {
            ids = new String[0];
            state = new ChunkState();
        }
    }

    private com.ooc.OB.ORBInstance orbInstance_;
    private com.ooc.CORBA.InputStream in_;
    private com.ooc.OCI.Buffer buf_;
    private java.util.Hashtable instanceTable_;
    private java.util.Hashtable headerTable_;
    private java.util.Hashtable positionTable_ = null;
    private ChunkState chunkState_ = new ChunkState();
    private Header currentHeader_ = null; // Java only

    // ------------------------------------------------------------------
    // Valuetype creation strategies
    // ------------------------------------------------------------------

    private static abstract class CreationStrategy
    {
        protected ValueReader reader_;
        protected com.ooc.CORBA.InputStream is_;

        CreationStrategy(ValueReader reader, com.ooc.CORBA.InputStream is)
        {
            reader_ = reader;
            is_ = is;
        }

        abstract java.io.Serializable
        create(Header h);
    }

    //
    // Create a valuebox using a BoxedValueHelper
    //
    private static class BoxCreationStrategy extends CreationStrategy
    {
        private org.omg.CORBA.portable.BoxedValueHelper helper_;

        BoxCreationStrategy(ValueReader reader,
                            com.ooc.CORBA.InputStream is,
                            org.omg.CORBA.portable.BoxedValueHelper helper)
        {
            super(reader, is);
            helper_ = helper;
        }

        java.io.Serializable
        create(Header h)
        {
            Assert._OB_assert(h.tag >= 0x7fffff00 && h.tag != -1);

            java.io.Serializable result = helper_.read_value(is_);

            if(result != null)
            {
                reader_.addInstance(h.headerPos, result);
                return result;
            }

            throw new org.omg.CORBA.MARSHAL(
                MinorCodes.describeMarshal(MinorCodes.MinorNoValueFactory) +
                    ": " + helper_.get_id(),
                MinorCodes.MinorNoValueFactory,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    //
    // Create a value using a class
    //
    private static class ClassCreationStrategy extends CreationStrategy
    {
        private Class clz_;

        ClassCreationStrategy(ValueReader reader,
                              com.ooc.CORBA.InputStream is,
                              Class clz)
        {
            super(reader, is);
            clz_ = clz;
        }

        java.io.Serializable
        create(Header h)
        {
            Assert._OB_assert(h.tag >= 0x7fffff00 && h.tag != -1);

            java.io.Serializable result = null;

            try
            {
                result = (java.io.Serializable)clz_.newInstance();
                reader_.addInstance(h.headerPos, result);
                try
                {
                    reader_.unmarshalValueState(result);
                }
                catch(org.omg.CORBA.SystemException ex)
                {
                    reader_.removeInstance(h.headerPos);
                    throw ex;
                }
                return result;
            }
            catch(ClassCastException ex)
            {
            }
            catch(IllegalAccessException ex)
            {
            }
            catch(InstantiationException ex)
            {
            }

            throw new org.omg.CORBA.MARSHAL(
                MinorCodes.describeMarshal(MinorCodes.MinorNoValueFactory) +
                    ": " + clz_.getName(),
                MinorCodes.MinorNoValueFactory,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    //
    // Create a value using a factory
    //
    private static class FactoryCreationStrategy extends CreationStrategy
    {
        private String id_;
        private ORBInstance orbInstance_;

        FactoryCreationStrategy(ValueReader reader,
                                com.ooc.CORBA.InputStream is,
                                String id)
        {
            super(reader, is);
            id_ = id;
            orbInstance_ = is._OB_ORBInstance();
        }

        private org.omg.CORBA.portable.ValueFactory
        findFactory(Header h, org.omg.CORBA.StringHolder id)
        {
            org.omg.CORBA.portable.ValueFactory f = null;

            if(orbInstance_ != null)
            {
                ValueFactoryManager manager =
                    orbInstance_.getValueFactoryManager();

                if(h.ids.length > 0)
                {
                    for(int i = 0 ; i < h.ids.length ; i++)
                    {
                        f = manager.lookupValueFactoryWithClass(h.ids[i]);
                        if(f != null)
                        {
                            id.value = h.ids[i];
                            break;
                        }

                        //
                        // If we have a formal ID, and we haven't found
                        // a factory yet, then give up
                        //
                        if(id_ != null && h.ids[i].equals(id_))
                            break;
                    }
                }
                else if(id_ != null)
                {
                    f = manager.lookupValueFactoryWithClass(id_);
                    id.value = id_;
                }
            }

            return f;
        }

        private java.io.Serializable
        createWithFactory(Header h,
                          org.omg.CORBA.portable.ValueFactory factory)
        {
            //
            // The factory's read_value method is expected to create
            // an instance of the valuetype and then call the method
            // InputStream.read_value(java.io.Serializable) to read
            // the valuetype state. We use a stack to remember the
            // Header information for use by initializeValue(), which
            // is called by read_value().
            //
            try
            {
                reader_.pushHeader(h);
                return factory.read_value(is_);
            }
            finally
            {
                reader_.popHeader();
            }
        }

        private org.omg.CORBA.portable.BoxedValueHelper
        getBoxedHelper(String id)
        {
            org.omg.CORBA.portable.BoxedValueHelper result = null;

            Class helperClass = Util.idToClass(id, "Helper");

            if(helperClass != null)
            {
                try
                {
                    return (org.omg.CORBA.portable.BoxedValueHelper)
                        helperClass.newInstance();
                }
                catch(ClassCastException ex)
                {
                }
                catch(IllegalAccessException ex)
                {
                }
                catch(InstantiationException ex)
                {
                }

                throw new org.omg.CORBA.MARSHAL(
                    MinorCodes.describeMarshal(
                        MinorCodes.MinorNoValueFactory) +
                        ": invalid BoxedValueHelper for " + id,
                    MinorCodes.MinorNoValueFactory,
                    org.omg.CORBA.CompletionStatus.COMPLETED_NO);
            }

            return null;
        }

        java.io.Serializable
        create(Header h)
        {
            org.omg.CORBA.StringHolder idH = new org.omg.CORBA.StringHolder();
            return create(h, idH);
        }

        java.io.Serializable
        create(Header h, org.omg.CORBA.StringHolder id)
        {
            Assert._OB_assert(h.tag >= 0x7fffff00 && h.tag != -1);

            //
            // See if a factory exists that can create the value
            //
            org.omg.CORBA.portable.ValueFactory factory = findFactory(h, id);
            if(factory != null)
                return createWithFactory(h, factory);

            //
            // Another possibility is that we're unmarshalling a valuebox,
            // so we'll try to load the Helper class dynamically
            //
            org.omg.CORBA.portable.BoxedValueHelper helper = null;
            if(h.ids.length > 0)
            {
                //
                // If it's a valuebox, at most one id will be marshalled
                //
                helper = getBoxedHelper(h.ids[0]);
                if(helper != null)
                    id.value = h.ids[0];
            }

            if(helper == null && id_ != null)
            {
                helper = getBoxedHelper(id_);
                if(helper != null)
                    id.value = id_;
            }

            if(helper != null)
            {
                java.io.Serializable result = helper.read_value(is_);
                reader_.addInstance(h.headerPos, result);
                return result;
            }

            String type = "<unknown>";
            if(h.ids.length > 0)
                type = h.ids[0];
            else if(id_ != null)
                type = id_;
            throw new org.omg.CORBA.MARSHAL(
                MinorCodes.describeMarshal(MinorCodes.MinorNoValueFactory) +
                    ": " + type,
                MinorCodes.MinorNoValueFactory,
                org.omg.CORBA.CompletionStatus.COMPLETED_NO);
        }
    }

    // ------------------------------------------------------------------
    // Private and protected members
    // ------------------------------------------------------------------

    // Java only
    private void
    addInstance(int pos, java.io.Serializable instance)
    {
        instanceTable_.put(new Integer(pos), instance);
    }

    // Java only
    private void
    removeInstance(int pos)
    {
        instanceTable_.remove(new Integer(pos));
    }

    private void
    readHeader(Header h)
    {
        //
        // Special cases are handled elsewhere
        //
        Assert._OB_assert(h.tag != 0 && h.tag != -1);

        //
        // Check if the value is chunked
        //
        if((h.tag & 0x00000008) == 8)
            h.state.chunked = true;

        //
        // Check for presence of codebase URL
        //
        if((h.tag & 0x00000001) == 1)
        {
            String codebase = null;

            //
            // Check for indirection tag
            //
            int save = buf_.pos_;
            int indTag = in_.read_long();
            if(indTag == -1)
            {
                int offs = in_.read_long();
                if(offs >= -4)
                    throw new org.omg.CORBA.MARSHAL();
                int tmp = buf_.pos_;
                buf_.pos_ = buf_.pos_ - 4 + offs;
                if(buf_.pos_ < 0)
                    throw new org.omg.CORBA.MARSHAL();
                codebase = in_.read_string();
                buf_.pos_ = tmp;
            }
            else
            {
                buf_.pos_ = save;
                codebase = in_.read_string();
            }
        }

        //
        // Extract repository ID information
        //
        if((h.tag & 0x00000006) == 0)
        {
            //
            // No type information was marshalled
            //
        }
        else if((h.tag & 0x00000006) == 6)
        {
            //
            // Extract a list of repository IDs, representing the
            // truncatable types for this value
            //

            //
            // Check for indirection tag (indirected list)
            //
            int saveList = buf_.pos_;
            int indTag = in_.read_long();
            boolean indList = (indTag == -1);

            if(indList)
            {
                int offs = in_.read_long();
                if(offs > -4)
                    throw new org.omg.CORBA.MARSHAL();
                saveList = buf_.pos_;
                buf_.pos_ = buf_.pos_ - 4 + offs;
                if(buf_.pos_ < 0)
                    throw new org.omg.CORBA.MARSHAL();
            }
            else
                buf_.pos_ = saveList;

            int count = in_.read_long();
            h.ids = new String[count];

            for(int i = 0 ; i < count ; i++)
            {
                //
                // Check for indirection tag (indirected list entry)
                //
                int saveRep = buf_.pos_;
                indTag = in_.read_long();
                if(indTag == -1)
                {
                    int offs = in_.read_long();
                    if(offs > -4)
                        throw new org.omg.CORBA.MARSHAL();
                    saveRep = buf_.pos_;
                    buf_.pos_ = buf_.pos_ - 4 + offs;
                    if(buf_.pos_ < 0)
                        throw new org.omg.CORBA.MARSHAL();
                    h.ids[i] = in_.read_string();
                    buf_.pos_ = saveRep;
                }
                else
                {
                    buf_.pos_ = saveRep;
                    h.ids[i] = in_.read_string();
                }
            }

            //
            // Restore buffer position (case of indirected list)
            //
            if(indList)
                buf_.pos_ = saveList;
        }
        else if((h.tag & 0x00000006) == 2)
        {
            //
            // Extract a single repository ID
            //
            String id;

            //
            // Check for indirection tag
            //
            int save = buf_.pos_;
            int indTag = in_.read_long();
            if(indTag == -1)
            {
                int offs = in_.read_long();
                if(offs > -4)
                    throw new org.omg.CORBA.MARSHAL();
                save = buf_.pos_;
                buf_.pos_ = buf_.pos_ - 4 + offs;
                if(buf_.pos_ < 0)
                    throw new org.omg.CORBA.MARSHAL();
                id = in_.read_string();
                buf_.pos_ = save;
            }
            else
            {
                buf_.pos_ = save;
                id = in_.read_string();
            }

            h.ids = new String[1];
            h.ids[0] = id;
        }

        //
        // Record beginning of value data
        //
        h.dataPos = buf_.pos_;

        //
        // Add entry to header table
        //
        headerTable_.put(new Integer(h.headerPos), h);
    }

    private void
    readChunk(ChunkState state)
    {
        //
        // Check for a chunk size
        //
        int size = in_._OB_readLongUnchecked();
        if(size >= 0 && size < 0x7fffff00) // chunk size
        {
            state.chunkStart = buf_.pos_;
            state.chunkSize = size;
        }
        else if(size < 0) // end tag
        {
            buf_.pos_ -= 4; // rewind
            state.chunkStart = buf_.pos_;
            state.chunkSize = 0;
        }
        else
        {
            buf_.pos_ -= 4; // rewind
            state.chunkStart = 0;
            state.chunkSize = 0;
        }
    }

    private void
    initHeader(Header h)
    {
        //
        // Null values and indirections must be handled by caller
        //
        Assert._OB_assert(h.tag != 0 && h.tag != -1);

        h.headerPos = buf_.pos_ - 4; // adjust for alignment
        h.state.copyFrom(chunkState_);

        //
        // Read value header info
        //
        readHeader(h);
        chunkState_.copyFrom(h.state);

        //
        // Increment our nesting level if we are chunked
        //
        if(chunkState_.chunked)
        {
            readChunk(chunkState_);
            chunkState_.nestingLevel++;
        }
    }

    private void
    skipChunk()
    {
        if(chunkState_.chunked)
        {
            //
            // At this point, the unmarshalling code has finished. However,
            // we may have a truncated value, or we may have unmarshalled a
            // custom value. In either case, we can't be sure that the
            // unmarshalling code has positioned the stream at the end of
            // this value.
            //
            // Therefore we will advance, chunk by chunk, until we reach
            // the end of the value.
            //

            //
            // Skip to the end of the current chunk (if necessary)
            //
            if(chunkState_.chunkStart > 0)
            {
                buf_.pos_ = chunkState_.chunkStart;
                in_._OB_skip(chunkState_.chunkSize);
            }

            chunkState_.chunkStart = 0;
            chunkState_.chunkSize = 0;

            //
            // Loop until we have reached the end of this value. We may
            // have to process nested values, chunks, etc.
            //
            int level = chunkState_.nestingLevel;
            int tag = in_._OB_readLongUnchecked();
            while(tag >= 0 || (tag < 0 && tag < -chunkState_.nestingLevel))
            {
                if(tag >= 0x7fffff00)
                {
                    //
                    // This indicates a nested value. We read the header
                    // information and store it away, in case a subsequent
                    // indirection refers to this value.
                    //
                    level++;
                    Header nest = new Header();
                    nest.tag = tag;
                    nest.headerPos = buf_.pos_ - 4; // adjust for alignment
                    nest.state.nestingLevel = level;
                    readHeader(nest);
                }
                else if(tag >= 0)
                {
                    //
                    // Chunk size - advance the stream past the chunk
                    //
                    in_._OB_skip(tag);
                }
                else
                {
                    //
                    // tag is less than 0, so this is an end tag for a nested
                    // value
                    //
                    level--;
                }

                //
                // Read the next tag
                //
                tag = in_._OB_readLongUnchecked();
            }

            //
            // If the tag is greater than our nesting level, then this
            // value coterminates with an outer value. We rewind the
            // stream so that the outer value can read this tag.
            //
            if(tag > -chunkState_.nestingLevel)
                buf_.pos_ -= 4;

            chunkState_.nestingLevel--;

            if(chunkState_.nestingLevel == 0)
                chunkState_.chunked = false;
            else
            {
                //
                // We're chunked and still processing nested values, so
                // another chunk may follow
                //
                readChunk(chunkState_);
            }
        }
    }

    //
    // Invoke the valuetype to unmarshal its state
    //
    // Java only
    //
    private void
    unmarshalValueState(java.io.Serializable v)
    {
        if(v instanceof org.omg.CORBA.portable.StreamableValue)
            ((org.omg.CORBA.portable.StreamableValue)v)._read(in_);
        else if(v instanceof org.omg.CORBA.CustomMarshal)
        {
            org.omg.CORBA.DataInputStream dis =
                new com.ooc.CORBA.DataInputStream(in_);
            ((org.omg.CORBA.CustomMarshal)v).unmarshal(dis);
        }
        else
            throw new org.omg.CORBA.MARSHAL("Valuetype does not implement " +
                                            "StreamableValue or " +
                                            "CustomMarshal");
    }

    private java.io.Serializable
    readIndirection(CreationStrategy strategy)
    {
        int offs = in_.read_long();
        int pos = buf_.pos_ - 4 + offs;
        pos += 3; // adjust for alignment
        pos -= (pos & 0x3);
        Integer posObj = new Integer(pos);

        //
        // Check the history for a value that was seen at the specified
        // position. Generally, we should expect the value to be present.
        // However, in the case of chunking, it's possible for an
        // indirection to refer to a nested value in the truncated state
        // of an enclosing value, or to a value that we could not
        // instantiate.
        //
        java.io.Serializable v =
            (java.io.Serializable)instanceTable_.get(posObj);

        if(v != null)
        {
            return v;
        }
        else
        {
            int save = buf_.pos_;

            //
            // Check for indirection to null value
            //
            buf_.pos_ = pos; // rewind to offset position
            if(in_._OB_readLongUnchecked() == 0)
            {
                buf_.pos_ = save;
                // Can't put a null in a Hashtable
                //instanceTable_.put(posObj, null);
                return null;
            }

            //
            // If it's not null and it's not in our history, then
            // there's no hope
            //
            Header nest = (Header)headerTable_.get(posObj);

            if(nest == null)
                throw new org.omg.CORBA.MARSHAL(
                    MinorCodes.describeMarshal(MinorCodes.MinorNoValueFactory)
                        + ": cannot instantiate value for indirection",
                    MinorCodes.MinorNoValueFactory,
                    org.omg.CORBA.CompletionStatus.COMPLETED_NO);

            //
            // Create the value
            //
            buf_.pos_ = nest.dataPos;
            ChunkState saveState = new ChunkState(chunkState_);
            chunkState_.copyFrom(nest.state);
            if(chunkState_.chunked)
                readChunk(chunkState_);

            try
            {
                v = strategy.create(nest);
            }
            finally
            {
                //
                // Restore state
                //
                buf_.pos_ = save;
                chunkState_.copyFrom(saveState);
            }

            return v;
        }
    }

    private java.io.Serializable
    read(CreationStrategy strategy)
    {
        Header h = new Header();
        h.tag = in_.read_long();

        if(h.tag == 0)
        {
            return null;
        }
        else if(h.tag == -1)
        {
            return readIndirection(strategy);
        }
        else if(h.tag < 0x7fffff00)
        {
            throw new org.omg.CORBA.MARSHAL(
                "Illegal valuetype tag 0x" + Integer.toHexString(h.tag));
        }
        else
        {
            initHeader(h);
            java.io.Serializable vb = strategy.create(h);
            skipChunk();
            return vb;
        }
    }

    //
    // Remarshal each valuetype member
    //
    private void
    copyValueState(org.omg.CORBA.TypeCode tc, com.ooc.CORBA.OutputStream out)
    {
        try
        {
            if(tc.kind() == org.omg.CORBA.TCKind.tk_value)
            {
                //
                // First copy the state of the concrete base type, if any
                //
                org.omg.CORBA.TypeCode base = tc.concrete_base_type();
                if(base != null)
                    copyValueState(base, out);

                for(int i = 0 ; i < tc.member_count() ; i++)
                    out.write_InputStream(in_, tc.member_type(i));
            }
            else if(tc.kind() == org.omg.CORBA.TCKind.tk_value_box)
            {
                out.write_InputStream(in_, tc.content_type());
            }
            else
                Assert._OB_assert(false);
        }
        catch(org.omg.CORBA.TypeCodePackage.BadKind ex)
        {
            Assert._OB_assert(false);
        }
        catch(org.omg.CORBA.TypeCodePackage.Bounds ex)
        {
            Assert._OB_assert(false);
        }
    }

    // Java only
    private void
    pushHeader(Header h)
    {
        h.next = currentHeader_;
        currentHeader_ = h;
    }

    // Java only
    private void
    popHeader()
    {
        Assert._OB_assert(currentHeader_ != null);

        currentHeader_ = currentHeader_.next;
    }

    //
    // Search up the valuetype's inheritance hierarchy for a TypeCode
    // with the given repository ID
    //
    private org.omg.CORBA.TypeCode
    findTypeCode(String id, org.omg.CORBA.TypeCode tc)
    {
        org.omg.CORBA.TypeCode result = null;
        org.omg.CORBA.TypeCode t = tc;
        while(result == null)
        {
            try
            {
                org.omg.CORBA.TypeCode t2 =
                    com.ooc.CORBA.TypeCode._OB_getOrigType(t);
                if(id.equals(t2.id()))
                {
                    result = t;
                }
                else if(t2.kind() == org.omg.CORBA.TCKind.tk_value &&
                        t2.type_modifier() ==
                        org.omg.CORBA.VM_TRUNCATABLE.value)
                {
                    t = t2.concrete_base_type();
                }
                else
                {
                    break;
                }
            }
            catch(org.omg.CORBA.TypeCodePackage.BadKind ex)
            {
                Assert._OB_assert(false);
            }
        }

        return result;
    }

    // ------------------------------------------------------------------
    // Public methods
    // ------------------------------------------------------------------

    public
    ValueReader(com.ooc.CORBA.InputStream in)
    {
        in_ = in;
        buf_ = in._OB_buffer();
        orbInstance_ = in._OB_ORBInstance();
        instanceTable_ = new java.util.Hashtable(131);
        headerTable_ = new java.util.Hashtable(131);
    }

    public java.io.Serializable
    readValue()
    {
        FactoryCreationStrategy strategy =
            new FactoryCreationStrategy(this, in_, null);
        return read(strategy);
    }

    public java.io.Serializable
    readValue(String id)
    {
        FactoryCreationStrategy strategy =
            new FactoryCreationStrategy(this, in_, id);
        return read(strategy);
    }

    public java.io.Serializable
    readValue(Class clz)
    {
        ClassCreationStrategy strategy =
            new ClassCreationStrategy(this, in_, clz);
        return read(strategy);
    }

    public java.io.Serializable
    readValueBox(org.omg.CORBA.portable.BoxedValueHelper helper)
    {
        BoxCreationStrategy strategy =
            new BoxCreationStrategy(this, in_, helper);
        return read(strategy);
    }

    public void
    initializeValue(java.io.Serializable value)
    {
        //
        // We should have previously pushed a Header on the stack
        //
        Assert._OB_assert(currentHeader_ != null);

        //
        // Now that we have an instance, we can put it in our history
        //
        addInstance(currentHeader_.headerPos, value);

        //
        // Unmarshal the value's state
        //
        try
        {
            unmarshalValueState(value);
        }
        catch(org.omg.CORBA.SystemException ex)
        {
            removeInstance(currentHeader_.headerPos);
            throw ex;
        }
    }

    public java.lang.Object
    readAbstractInterface()
    {
        //
        // Abstract interfaces are marshalled like a union with a
        // boolean discriminator - if true, an objref follows,
        // otherwise a valuetype follows
        //
        if(in_.read_boolean())
            return in_.read_Object();
        else
            return readValue();
    }

    public java.lang.Object
    readAbstractInterface(Class clz)
    {
        //
        // Abstract interfaces are marshalled like a union with a
        // boolean discriminator - if true, an objref follows,
        // otherwise a valuetype follows
        //
        if(in_.read_boolean())
            return in_.read_Object(clz);
        else
            return readValue(clz);
    }

    //
    // Copy a value from out InputStream to the given OutputStream.
    // The return value is the most-derived TypeCode of the value
    // written to the stream. This may not be the same as the TypeCode
    // argument if the value is truncated. Furthermore, the TypeCode
    // may not represent the actual most-derived type, since a more-derived
    // value may have been read.
    //
    public org.omg.CORBA.TypeCode
    remarshalValue(org.omg.CORBA.TypeCode tc, com.ooc.CORBA.OutputStream out)
    {
	//
	// TODO: We've removed the reset of the position table at each top
	// level call.  We need to perform more testing and analysis to verify
	// that this wasn't broken.  remarshalDepth was also removed since it
	// wasn't being used anywhere except to determine when the table should
	// be reset.
	//

        //
        // Create a new Hashtable for each top-level call to remarshalValue
        //
        if(positionTable_ == null)
            positionTable_ = new java.util.Hashtable(131);

        org.omg.CORBA.TypeCode origTC =
            com.ooc.CORBA.TypeCode._OB_getOrigType(tc);

        org.omg.CORBA.TypeCode result = null;

	Header h = new Header();
	h.tag = in_.read_long();
	h.headerPos = buf_.pos_ - 4; // adjust for alignment
	h.state.copyFrom(chunkState_);

	//
	// Remember starting position of this valuetype
	//
	int pos = h.headerPos;

	//
	// Check for special cases (null values and indirections)
	//
	if(h.tag == 0)
	{
	    out.write_long(0);
	    result = tc;
	}
	else if(h.tag == -1)
	{
	    //
	    // Since offsets can change as we remarshal valuetypes to the
	    // output stream, we use a table to map old positions to new ones
	    //
	    int offs = in_.read_long();
	    int oldPos = buf_.pos_ - 4 + offs;
	    oldPos += 3; // adjust alignment to start of value
	    oldPos -= (oldPos & 0x3);

	    //
	    // If we find the position in the table, write the translated
	    // offset to the output stream. Otherwise, the indirection refers
	    // to a valuetype that we were unable to create and we therefore
	    // raise MARSHAL.
	    //
	    Integer newPos =
		(Integer)positionTable_.get(new Integer(oldPos));
	    if(newPos != null)
	    {
		out.write_long(h.tag);
		offs = newPos.intValue() - out._OB_pos();
		out.write_long(offs);
		//
		// TODO: The TypeCode may not necessarily reflect the
		// TypeCode of the indirected value.
		//
		result = tc;
	    }
	    else
		throw new org.omg.CORBA.MARSHAL(
		    "Cannot find value for indirection");
	}
	else
	{
	    //
	    // Add valuetype to position map
	    //
	    int outPos = out._OB_pos();
	    outPos += 3; // adjust alignment to start of value
	    outPos -= (outPos & 0x3);
	    positionTable_.put(new Integer(pos), new Integer(outPos));

	    //
	    // Read value header info
	    //
	    readHeader(h);
	    chunkState_.copyFrom(h.state);

	    if(chunkState_.chunked)
	    {
		readChunk(chunkState_);
		chunkState_.nestingLevel++;
	    }

	    String tcId = null;
	    short mod = org.omg.CORBA.VM_NONE.value;
	    int idPos;
	    try
	    {
		tcId = origTC.id();
		if(origTC.kind() == org.omg.CORBA.TCKind.tk_value)
		    mod = origTC.type_modifier();
	    }
	    catch(org.omg.CORBA.TypeCodePackage.BadKind ex)
	    {
		Assert._OB_assert(false);
	    }

	    //
	    // We have two methods of extracting the state of a valuetype:
	    //
	    // 1) Use the TypeCode
	    // 2) Use a valuetype factory
	    //
	    // Which method we use is determined by the repository IDs
	    // in the valuetype header.
	    //
	    // Our goal is to preserve as much information as possible.
	    // If the TypeCode describes a more-derived type than any
	    // available factory, we will use the TypeCode. Otherwise,
	    // we use a factory to create a temporary instance of the
	    // valuetype, and subsequently marshal that instance to the
	    // OutputStream.
	    //

	    //
	    // See if the TypeCode ID matches any of the valuetype's IDs -
	    // stop at the first match
	    //
	    String id = null;
	    for(idPos = 0 ; idPos < h.ids.length ; idPos++)
		if(tcId.equals(h.ids[idPos]))
		{
		    id = h.ids[idPos];
		    break;
		}

	    //
	    // See if a factory exists for any of the valuetype's IDs -
	    // stop at the first match
	    //
	    String factoryId = null;
	    int factoryPos = 0;
	    org.omg.CORBA.portable.ValueFactory factory = null;
	    if(orbInstance_ != null)
	    {
		ValueFactoryManager manager =
		    orbInstance_.getValueFactoryManager();

		for(factoryPos = 0 ; factoryPos < h.ids.length ;
		    factoryPos++)
		{
		    factory = manager.lookupValueFactoryWithClass(
			h.ids[factoryPos]);
		    if(factory != null)
		    {
			factoryId = h.ids[factoryPos];
			break;
		    }
		}
	    }

	    //
	    // If no ID matched the TypeCode, and no factory was found,
	    // then we have no way to remarshal the data
	    //
	    if(h.ids.length > 0 && id == null && factoryId == null)
		throw new org.omg.CORBA.MARSHAL(
		    MinorCodes.describeMarshal(
			MinorCodes.MinorNoValueFactory) +
			": insufficient information to copy valuetype",
		    MinorCodes.MinorNoValueFactory,
		    org.omg.CORBA.CompletionStatus.COMPLETED_NO);

	    //
	    // If value is custom and there is no factory, then we have
	    // no way to remarshal the data
	    //
	    if(mod == org.omg.CORBA.VM_CUSTOM.value && factoryId == null)
		throw new org.omg.CORBA.MARSHAL(
		    MinorCodes.describeMarshal(
			MinorCodes.MinorNoValueFactory) +
			": unable to copy custom valuetype",
		    MinorCodes.MinorNoValueFactory,
		    org.omg.CORBA.CompletionStatus.COMPLETED_NO);

	    //
	    // If the TypeCode is more descriptive than any available
	    // factory, or if no identifiers were provided, then use the
	    // TypeCode, otherwise use the factory
	    //
	    // NOTE: (Java only) We also use the TypeCode for boxed values,
	    //       because we don't have the BoxedHelper and may not be
	    //       able to locate one via the class loader.
	    //
	    if(idPos < factoryPos || h.ids.length == 0 ||
	       origTC.kind() == org.omg.CORBA.TCKind.tk_value_box)
	    {
		int i;

		//
		// We may need to truncate the state of this value, which
		// means we need to revise the list of repository IDs
		//
		int numIds = h.ids.length - idPos;
		String[] ids = new String[numIds];
		for(i = idPos ; i < h.ids.length ; i++)
		    ids[i - idPos] = h.ids[i];

		out._OB_beginValue(h.tag, ids, h.state.chunked);
		copyValueState(origTC, out);
		out._OB_endValue();

		result = tc;
	    }
	    else
	    {
		//
		// Create a temporary instance to use for marshalling
		//
		try
		{
		    pushHeader(h);
		    java.io.Serializable vb = factory.read_value(in_);
		    try
		    {
			out.write_value(vb);
		    }
		    finally
		    {
			removeInstance(h.headerPos);
		    }
		}
		finally
		{
		    popHeader();
		}

		//
		// Determine the TypeCode that is equivalent to the
		// factory in use
		//
		result = findTypeCode(h.ids[factoryPos], tc);

		if(result == null)
		    result = tc;
	    }

	    skipChunk();
	}

        Assert._OB_assert(result != null);
        return result;
    }

    public void
    readValueAny(org.omg.CORBA.Any any, org.omg.CORBA.TypeCode tc)
    {
        //
        // In constrast to other complex types, valuetypes and value boxes
        // in Anys cannot simply be "remarshalled". The reason is that
        // indirection may occur across multiple Any values in the same
        // stream. Therefore, if possible, we should attempt to construct
        // the values using a factory so that any shared values will be
        // handled correctly.
        //
        // If a factory cannot be found, we should still remarshal.
        // However, if an indirection is encountered which refers to
        // a value we were unable to construct, an exception will be
        // raised.
        //

        com.ooc.CORBA.Any obAny = null;
        try
        {
            obAny = (com.ooc.CORBA.Any)any;
        }
        catch(ClassCastException ex)
        {
            //
            // Any may have been created by a foreign ORB
            //
        }

        org.omg.CORBA.TypeCode origTC =
            com.ooc.CORBA.TypeCode._OB_getOrigType(tc);

        //
        // Check if the Any contains an abstract interface
        //
        if(origTC.kind() == org.omg.CORBA.TCKind.tk_abstract_interface)
        {
            boolean b = in_.read_boolean();
            if(b)
            {
                //
                // The abstract interface represents an object reference
                //
                any.insert_Object(in_.read_Object(), tc);
                return;
            }
            else
            {
                //
                // The abstract interface represents a valuetype. The
                // readValue() method will raise an exception if an
                // instance of the valuetype could not be created.
                // We cannot remarshal in this case because we don't
                // have a TypeCode for the valuetype.
                //
                any.insert_Value(readValue(), tc);
                return;
            }
        }

        //
        // If the TypeCode's repository ID is that of CORBA::ValueBase,
        // then we try to create an instance. The Any could contain a
        // valuetype *or* a boxed valuetype.
        //
        // If creation fails, we cannot remarshal the value, since
        // CORBA::ValueBase is not a truncatable base type, and we
        // have no other TypeCode information. Truncating to
        // CORBA::ValueBase doesn't seem very useful anyway.
        //
        try
        {
            String id = origTC.id();
            if(id.equals("IDL:omg.org/CORBA/ValueBase:1.0"))
            {
                any.insert_Value(readValue(), tc);
                return;
            }
        }
        catch(org.omg.CORBA.TypeCodePackage.BadKind ex)
        {
            Assert._OB_assert(false);
        }

        //
        // At this point, the Any contains a valuetype or a boxed valuetype,
        // and we have a TypeCode that can be used for remarshalling.
        //

        //
        // Save some state so that we can restore things prior to
        // remarshalling
        //
        int startPos = buf_.pos_;
        ChunkState startState = new ChunkState(chunkState_);

        //
        // No need to worry about truncation for boxed valuetypes
        //
        if(origTC.kind() == org.omg.CORBA.TCKind.tk_value_box)
        {
            try
            {
                any.insert_Value(readValue(tc.id()), tc);
                return;
            }
            catch(org.omg.CORBA.MARSHAL ex)
            {
                //
                // Creation failed - restore our state and try remarshalling
                //
                buf_.pos_ = startPos;
                chunkState_.copyFrom(startState);

                com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
                com.ooc.CORBA.OutputStream out =
                    new com.ooc.CORBA.OutputStream(buf);
                out._OB_ORBInstance(orbInstance_);
                remarshalValue(origTC, out);
                com.ooc.CORBA.InputStream in =
                    (com.ooc.CORBA.InputStream)out.create_input_stream();
                Assert._OB_assert(obAny != null);
                obAny.replace(tc, in);
            }
            catch(org.omg.CORBA.TypeCodePackage.BadKind ex)
            {
                Assert._OB_assert(false);
            }
        }
        else
        {
            //
            // Read valuetype header tag
            //
            Header h = new Header();
            h.tag = in_.read_long();

            //
            // Check tag for special cases
            //
            if(h.tag == 0)
            {
                any.insert_Value(null, tc);
                return;
            }
            else if(h.tag != -1 && h.tag < 0x7fffff00)
            {
                throw new org.omg.CORBA.MARSHAL(
                    "Illegal valuetype tag 0x" + Integer.toHexString(h.tag));
            }

            //
            // Try to create an instance of the valuetype using a factory
            //
            FactoryCreationStrategy strategy =
                new FactoryCreationStrategy(this, in_, null);
            try
            {
                if(h.tag == -1)
                {
                    //
                    // TODO: The TypeCode may not necessarily reflect
                    // the one that was used for the indirected value
                    // (i.e., the value may have been truncated).
                    // Fixing this probably requires maintaining a
                    // map of stream position to TypeCode.
                    //
                    any.insert_Value(readIndirection(strategy), tc);
                    return;
                }
                else
                {
                    initHeader(h);
                    org.omg.CORBA.StringHolder idH =
                        new org.omg.CORBA.StringHolder();
                    java.io.Serializable vb = strategy.create(h, idH);
                    skipChunk();

                    //
                    // Find the TypeCode for the repository ID that was
                    // used to instantiate the valuetype. Three things
                    // can happen:
                    //
                    // 1) The TypeCode is equal to tc.
                    //
                    // 2) The TypeCode is null. In this case, the instance
                    //    is of a more-derived type than tc, so the best
                    //    we can do is to use tc.
                    //
                    // 3) The TypeCode is a base type of tc. In this case,
                    //    the valuetype was truncated.
                    //
                    org.omg.CORBA.TypeCode t = findTypeCode(idH.value, tc);
                    if(t != null)
                        any.insert_Value(vb, t);
                    else
                        any.insert_Value(vb, tc);
                    return;
                }
            }
            catch(org.omg.CORBA.MARSHAL ex)
            {
                //
                // Creation failed - restore our state and try remarshalling
                //
                buf_.pos_ = startPos;
                chunkState_.copyFrom(startState);

                com.ooc.OCI.Buffer buf = new com.ooc.OCI.Buffer();
                com.ooc.CORBA.OutputStream out =
                    new com.ooc.CORBA.OutputStream(buf);
                out._OB_ORBInstance(orbInstance_);
                org.omg.CORBA.TypeCode t = remarshalValue(origTC, out);
                com.ooc.CORBA.InputStream in =
                    (com.ooc.CORBA.InputStream)out.create_input_stream();
                Assert._OB_assert(obAny != null);
                obAny.replace(t, in);
                return;
            }
        }

        Assert._OB_assert(false); // should never reach this point
    }

    public void
    beginValue()
    {
        Header h = new Header();
        h.tag = in_.read_long();
        Assert._OB_assert(h.tag != 0 && h.tag != -1);

        initHeader(h);
    }

    public void
    endValue()
    {
        skipChunk();
    }

    public void
    checkChunk()
    {
        if(!chunkState_.chunked)
            return;

        //
        // If we've reached the end of the current chunk, then check
        // for the start of a new chunk
        //
        if(chunkState_.chunkStart > 0 &&
           chunkState_.chunkStart + chunkState_.chunkSize == buf_.pos_)
            readChunk(chunkState_);
    }
}
